#!/usr/bin/perl  -w

# RSS Reader
# Copyright (C) 2007  Stephen Ostermiller
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;
# sudo perl -MCPAN -e 'install XML::RSS'
use XML::RSS;
use HTML::Scrubber;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Response;
use Time::Local;
use CGI qw/:standard/;
use XML::XPath;
use MIME::Base64;
use Unicode::String;
use threads;

my $debug = 0;

my @param = param();

my $totalItemCount=0;
my $totalFeedCount=0;
my ($page) = &alphaNumOnly(param('page'));

if ($ENV{'SERVER_NAME'}){
    print "Content-type: text/html; charset=UTF-8\n";
    print "\n";
}

if ($page eq ""){
    &printIndex();
    exit;
}

my $xp = XML::XPath->new(filename => "$page.page.xml");
my $pagenode = $xp->find('/page')->get_node(1);

my $pagetitle=$pagenode->find('@title')->string_value;
if (!$pagetitle){
    $pagetitle = "News RSS";
}

my $pagedescription=$pagenode->find('@description')->string_value;
if (!$pagedescription){
    $pagedescription = "News aggregated from RSS feeds";
}

print <<ZZZHEAD;
<html>
<head>
<title>$pagetitle</title>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<meta name="robots" content="noindex,nofollow,noarchive,none">
<meta name="description" content="$pagedescription">
<link rel=StyleSheet href="style.css" type="text/css">
</head>
<body>
ZZZHEAD

&updateAllRSS($xp);
&showAllRSS($xp, $page);

print <<ZZZFOOT;
</body>
</html>
ZZZFOOT

sub toBooleanString(){
    my ($s) = @_;
    if (&isAffirmative($s,0)){
        return "true";
    }
    return "false";
}

sub updateAllRSS(){
    my ($xp) = @_;
    print "<div id=updates>\n";
    my @thr = ();
    my $count = 0;
    foreach my $feed ($xp->find('//feed')->get_nodelist){
        my $url = $feed->find('@url')->string_value;
        my $title = $feed->find('@feedTitle')->string_value;
        my $cacheMinutes = $feed->find('@cacheMinutes')->string_value;
        push(@thr, threads->new(\&updateRSS, $url, $title, $count, $cacheMinutes));
        $count++;
    }    
    my $success = 1;    
    foreach my $t (@thr){
        $success = $success & $t->join();
    }
    print "</div>\n";
    if ($success){
        print "<script>setTimeout('document.getElementById(\"updates\").innerHTML=\"\"',3000);</script>\n";
    }
}

sub fixupUrl(){
    my ($url) = @_;
     my $basehref = "";
    if ($ENV{'SERVER_NAME'}){
        my $uri = $ENV{'SCRIPT_NAME'};
        $uri =~ s/index\.cgi$//g;  
        $uri =~ s/\/$//g;
        my $port = $ENV{'SERVER_PORT'};
        if (param("port") and param("port") =~ /^\d+$/){
            $port = param("port");
        }
        $basehref = "http://$ENV{'SERVER_NAME'}:$port$uri";
    }
    if ($url =~ /^\//){
        $url = "$basehref$url";
    }
    return $url;
}

sub updateRSS(){
    my ($url, $title, $count, $cacheMinutes) = @_;    
    $url = &fixupUrl($url);
    
    my $file = &getRssLocalFileName($url);
    my $lastModifiedMinutes = -1;
    if (-e $file){
        $lastModifiedMinutes = -M $file;
    }
    $lastModifiedMinutes = int($lastModifiedMinutes * 24 * 60);
    if (!$cacheMinutes or  $lastModifiedMinutes < 0 or $lastModifiedMinutes > $cacheMinutes){
        my $urlsafe = $url;
        $urlsafe =~ s/^(http\:\/\/[^\/]+\/).*/$1.../g;
        my $display = $title;
        if (!$display or $display eq ""){
            $display = $urlsafe;
        }
        my $id="feedind".$count;

        print "<div class=updating id=$id>Updating ",escapeHTML($display),"</div>\n";
        &flush();
        my $ua = new LWP::UserAgent();
        $ua->timeout(10);           
        my $request = new HTTP::Request('GET', $url);
        my $response = $ua->request($request);
        if (!$response->is_success){
            print "<script>document.getElementById(\"$id\").innerHTML+=\" <span class=error>Could not update</span>\";</script>\n";
            return 0;
        }
        my $doc = $response->content;
        if ($doc eq ""){
            print "<script>document.getElementById(\"$id\").innerHTML+=\" <span class=error>Got empty document</span>\";</script>\n";
            return 0;
        }
        if (!open(FILEOUT, ">$file")){
            print "<script>document.getElementById(\"$id\").innerHTML+=\" <span class=error>Could not save to file</span>\";</script>\n";
            return 0;
        }
        print FILEOUT $doc;
        close(FILEOUT);
        print "<script>document.getElementById(\"$id\").innerHTML=\"\";</script>\n";
        &flush();
    }
    return 1;
}

sub showAllRSS(){
    my ($xp, $page) = @_;
    print "<table border=0><tr>\n";
    my $columnnum = 0;
    foreach my $column ($xp->find('/page/column')->get_nodelist){
        print "<td valign=top>\n";
        my $rownum = 0;
        foreach my $elementnode ($column->find('*')->get_nodelist){
            if ($elementnode->getName() eq "group"){
                my @rssitems = ();
                my @heads = ();
                foreach my $feed ($elementnode->find('feed')->get_nodelist){
                    my ($head, @subitems) = getRSS($feed, $elementnode);
                    push(@rssitems, @subitems);
                    push(@heads, $head);
                }
                &printRSS($elementnode, \@heads, \@rssitems);                
            } elsif ($elementnode->getName() eq "feed"){
                my ($head, @rssitems) = getRSS($elementnode, undef);
                my @heads = ($head);
                &printRSS(undef, \@heads, \@rssitems);
            }
            $rownum++;
        }
        print "</td>\n";
        $columnnum++;
    }
    print "</tr></table>\n";
}

sub getExistingValue(){
    my ($feed, $value) = @_;
    return &ensureExists($feed->find('@'.$value)->string_value())    
}

sub ensureExists(){
    my ($s, $r) = @_;
    if (!$s){
        if (!$r){
            $r = "";
        }
        return "$r";
    }
    return $s;
}

sub ensureExistsZero(){
    my ($s) = @_;
    if (!$s){
        return 0;
    }
    return $s;
}

sub getRssLocalFileName(){
    my ($url) = @_;
    if (!$url){
        return "";
    }
    $url = &fixupUrl($url);

    my $name;
    if ($url =~ /^http\:\/\/([a-zA-Z0-9\.]+)(.+)$/){
        my $site = $1;
        my $path = $2;
        $name = $site."_".encode_base64($path);
    } else {
        $name = encode_base64($url);
    }
    $name =~ s/[\r\n \t\/]//g;
    return "cache_".$name.".rss";
}

sub isAffirmative(){
    my ($s, $default) = @_;
    if (!$s or $s eq ""){
        return $default;
    }
    if ($s =~ /^yes|y|full|on|true|t|(?:[0-9]*[1-9][0-9]*)|sure|affirmative|yeah|absolutely$/i){
        return 1;
    }
    return 0;
}

sub printRSS(){
    my ($group, $headsref, $itemsref) = @_;
    
    my @heads = @$headsref;
    
    my @rssitems = @$itemsref;
    
    my $sort;
    my $itemCount;
    if ($group){
        $sort = $group->find('@sort')->string_value();
        $itemCount = $group->find('@itemCount')->string_value();
        my $groupName = $group->find('@groupName')->string_value();
        my $groupDescription = $group->find('@groupDescription')->string_value();
        my $groupLink = $group->find('@groupLink')->string_value();
        if ($groupDescription){
            $groupDescription = "<div class=groupDescription>$groupDescription</div>\n"; 
        } else {
            $groupDescription = "";
        }
        if ($groupName){
            @heads = (
                "<h2 class=groupTitle>".&formatDisplayLink($groupLink, $groupName)."</h2>\n$groupDescription"
            )
        }
    }

    print "<div class=feed>\n";
    foreach my $head (@heads){
        print $head;
    }

    my $itemIndex = 0;

    @rssitems = &sortRSS($sort, @rssitems);
        
    my %seenurls = ();

    foreach my $rssitem (@rssitems) {
        my $itemlink = $rssitem->getLink();
        
        if (!$seenurls{$itemlink} and (!$itemCount or $itemIndex < $itemCount)){            
            if ($rssitem->passesFilter()){

                my $evenodd = ($totalItemCount % 2 == 1) ? "odd" : "even";                    
                print "<div class=\"item $evenodd\">\n";

                print $rssitem->getDisplayTitle();

                print $rssitem->getDisplaySource();

                print $rssitem->getDisplayAge();

                print $rssitem->getDisplayDescription();

                print $rssitem->getDisplayHitParade();

                print $rssitem->getDisplayDiscussLink();

                print "</div>\n";
                $itemIndex++;
                $totalItemCount++;
            }
        }
        $seenurls{$itemlink} = 1;
    }

    print "</div>\n";
    &flush();
}

sub sortRSS(){
    my ($sort, @rssitems) = @_;
    if (!$sort or $sort eq ''){
        # no sort
    } elsif ($sort eq "date"){
        @rssitems = sort { ($b->getPublishedTime()) <=> ($a->getPublishedTime()) } @rssitems;
    } elsif ($sort eq "rating"){
        @rssitems = sort { ($b->getRating()) <=> ($a->getRating()) } @rssitems;
    } else {
        die "unrecognized sort: $sort";
    }
    return @rssitems;
}

sub getRSS(){
    my ($feed, $group) = @_;
    my $baseitem = rssitem(0,$feed,$group,0);
    my $url = &getExistingValue($feed, "url");    
    my $itemCount = &getExistingValue($feed, "itemCount");
    my $feedtitle = &getExistingValue($feed, "feedTitle");
    my $feeddescription = &getExistingValue($feed, "feedDescription");
    my $file = &getRssLocalFileName($url);
    
    my $head = "";
    my @rssitems = ();

    eval {
        my $rss = new XML::RSS;
        my $lastModifiedMinutes = 1;
        if (! -e $file){
            die "$file does not exist";
        }
        $lastModifiedMinutes = -M $file;
        my $rssxml = &readXML($file);
        $rss->parse($rssxml);
        $lastModifiedMinutes = int($lastModifiedMinutes * 24 * 60);
        my $min = $lastModifiedMinutes%60;
        my $hour = int($lastModifiedMinutes/60);
        my $old = "";
        if ($lastModifiedMinutes == 0){
            $old = "just updated";
        } else {
            if ($min == 1){
                $old = "1 minute";
            } elsif ($min > 1){
                $old = "$min minutes";
            }
            if ($hour == 1){
                $old = "1 hour ".$old;
            } elsif ($hour > 1){
                $old = "$hour hours ".$old;
            }
            $old = $old." old";
        }
        
        $feedtitle = $rss->channel('title') if ($feedtitle eq "");
        $feeddescription = $rss->channel('description') if ($feedtitle eq "");
        my $feedlink = $rss->channel('link');

        $head = "<h2 class=feedtitle>".&formatDisplayLink($feedlink, $feedtitle)."</h2>\n";
        if ($baseitem->showFeedDescription() and $feeddescription){
            $head.="<div class=feeddescription>$feeddescription</div>\n";
        }

        my @items = @{$rss->{'items'}};
        foreach my $item (@items){
            my $rssitem = rssitem($item,$feed,$group,$rss);
            push(@rssitems, $rssitem);
        }
        
        my $sort = &ensureExists($feed->find('@sort')->string_value);
        @rssitems = &sortRSS($sort, @rssitems);
        
        if ($itemCount and $itemCount-1 < $#rssitems){
            $#rssitems = $itemCount-1;
        }
    };
    if ($@){
        print STDERR $@;
        $head = (
            "<div class=error>Could not parse RSS: ".escapeHTML(&truncate($url,30))."</div>\n"
        );
        @rssitems = ();
    }

    return $head, @rssitems;
}

sub passesFilter(){
    my ($rssitem) = @_;
    my $itemtitle = $rssitem->getTitle();
    my $itemdescription = $rssitem->getDescription();
    my $filter = $rssitem->filter();
    if ($filter and $filter ne "" and ($itemtitle =~ /\b$filter\b/i or $itemdescription =~ /\b$filter\b/i)){
        return 0;
    }
    return 1;
}

sub getPublishedTime(){
    my ($rssitem) = @_;
    if ($rssitem->{TIME} == 0){
        $rssitem->{TIME} = &parseDate($rssitem->getPublishedDate());
    }
    return $rssitem->{TIME};
}

sub getPublishedDate(){
    my ($rssitem) = @_;
    my $publishedDate = $rssitem->item()->{'pubDate'};
    if (!$publishedDate){
        $publishedDate = $rssitem->item()->{'http://purl.org/dc/elements/1.1/'}->{date};
    }
    return $publishedDate;
}

sub getRating(){
    my ($rssitem) = @_;
    my $rating = $rssitem->item()->{'http://www.purl.org/stuff/rev/'}->{rating};
    if ($rating){
        return $rating;
    }
    return 0;
}

sub getFullDescription(){
    my ($rssitem) = @_;
    my $itemdescription = $rssitem->item->{'description'};
    if (!$itemdescription){
        $itemdescription=$rssitem->item->{'http://www.w3.org/2005/Atom'}->{'summary'};
    }    
    my $content = $rssitem->item->{'http://purl.org/rss/1.0/modules/content/'}->{encoded};
    if ($content){
        if (&startTheSame($content, $itemdescription)){
            $itemdescription = $content;
        } else {
            $itemdescription = "$itemdescription<p>\n$content";
        }
    }
    if (!$itemdescription){
        $itemdescription = '';
    }
    return $itemdescription;
}

sub getDescription(){
    my ($rssitem) = @_;
    my $itemdescription = $rssitem->getFullDescription();
    my $strip = $rssitem->strip();
    if ($strip){
        $itemdescription =~ s/$strip//gsi;
    }
    return $itemdescription;
}

sub getDisplayDescription(){
    my ($rssitem) = @_;
    my $itemdescription = $rssitem->getDescription();
    my $displaydesc = "";
    my $showDescription = $rssitem->showDescription();
    if (&isAffirmative($showDescription,1) and $itemdescription ne ""){
        if ($showDescription =~ /^[0-9]+$/){
            $itemdescription = &toSafeHTML($itemdescription, 0, 0, 0);
            $itemdescription = &truncate($itemdescription, $showDescription);
        } else {
            $itemdescription = &toSafeHTML($itemdescription, $rssitem->allowFormatting(), $rssitem->allowLinks(), $rssitem->allowImages());
        }
        $displaydesc .= "<div class=description>\n";
        $displaydesc .= "$itemdescription\n";
        $displaydesc .= "</div>\n";
    }
    return $displaydesc;
}

sub getTitle(){
    my ($rssitem) = @_;
    return &toSafeHTML($rssitem->item->{'title'}, 0, 0, 0);
}

sub getDisplayTitle(){
    my ($rssitem) = @_;
    my $itemtitle = $rssitem->getTitle();
    my $itemlink = $rssitem->getLink();
    if ($rssitem->showTitle()){
        return "<h3 class=itemtitle>".&formatDisplayLinkUnescaped($itemlink,$itemtitle)."</h3>\n";
    }
    return "";
}


sub getDisplayDiscussLink(){
    my ($rssitem) = @_;
    my $itemlink = $rssitem->getOriginalLink(); 
    if ($itemlink and $itemlink ne "" and $rssitem->addDiscussLink()){
        return "<p>".&formatDisplayLink($itemlink,"discuss...")."</p>\n";
    }
    return "";
}

sub getHitParade(){
    my ($rssitem) = @_;  
    return $rssitem->item()->{'http://purl.org/rss/1.0/modules/slash/'}->{hit_parade};
}

sub getDisplayHitParade(){
    my ($rssitem) = @_;  
    my $hitParade = $rssitem->getHitParade();
    if ($hitParade){
        my $itemlink = $rssitem->getLink();
        my $bestthreshold = -1;
        my $bestcomments = 0;
        my @hits = split(/,/, $hitParade);
        my $threshold=-1;
        foreach my $hit(@hits){
            if ($bestcomments < 10 && $hit > $bestcomments
                    || $bestcomments >= 10 && $hit >= 10 && $hit < $bestcomments){
                $bestcomments = $hit;
                $bestthreshold = $threshold
            }
            $threshold++;
        }
        return "<p class=comments>".&formatDisplayLink("$itemlink&threshold=$bestthreshold", "$bestcomments Comments (Threshold: $bestthreshold)")."</p>\n";
    }
    return "";
}

sub getDisplayAge(){
    my ($rssitem) = @_;
    my $displayAge = &getAgeString($rssitem->getPublishedTime());
    if ($displayAge ne ""){
        return "<div class=age>$displayAge</div>";
    }
    return "";
}

sub getDisplaySource(){
    my ($rssitem) = @_;
    my $source = &escapeHTML($rssitem->source());
    if ($rssitem->showSource() and $source){
        $source = &formatDisplayLink($rssitem->getFeedLink(), $source);
        return "<div class=source>$source</div>";
    }
    return "";
}

sub formatDisplayLinkUnescaped(){
    my ($link, $text) = @_;
    if ($link){
        return "<a href=\"".&escapeHTML($link)."\">$text</a>";
    }
    return $text;
}

sub formatDisplayLink(){
    my ($link, $text) = @_;
    return &formatDisplayLinkUnescaped($link,&escapeHTML($text));
}

sub getFeedLink(){
    my ($rssitem) = @_;
    return $rssitem->rss()->channel('link');
}

sub getOriginalLink(){
    my ($rssitem) = @_;
    my $link = $rssitem->item->{'link'};    
    if (!$link){
        $link = '';
    }
    return $link;
}

sub getLink(){
    my ($rssitem) = @_;
    my $link = $rssitem->getOriginalLink();
    my $itemdescription = $rssitem->getFullDescription();
    if ($rssitem->urlInDescription() and $itemdescription =~ /(http\:\/\/[^"' ]+)/){
        $link = $1;
    }
    return $link;
}

sub startTheSame(){
    my ($s1, $s2) = @_;
    $s1 =~ s/\<.*?\>//g;
    $s1 =~ s/[^A-Za-z0-9]//g;
    $s2 =~ s/\<.*?\>//g;
    $s2 =~ s/[^A-Za-z0-9]//g;
    if (substr($s1,0,20) eq substr($s2,0,20)){
        return 1;
    } else {
        return 0;
    }
}

sub truncate(){
    my ($s, $length) = @_;
    if (length($s) > $length){
        $s = substr($s,0,$length);
        $s =~ s/[^A-Za-z0-9][A-Za-z0-9]*$/.../gs;
    }
    return $s;
}

sub toSafeHTML(){
    my ($s, $allowFormatting, $allowLinks, $allowImages) = @_;

    my @scrubberAllow = ();
    my @scrubberRules = ();

    if (&isAffirmative($allowFormatting, 0)){
        push(@scrubberAllow, "br");
        push(@scrubberAllow, "p");
        push(@scrubberAllow, "b");
        push(@scrubberAllow, "i");
        push(@scrubberAllow, "u");
    }

    if (&isAffirmative($allowLinks, 0)){
        push(@scrubberAllow, "a");
        push(
            @scrubberRules,
            a => {
                href => qr{^http[s]?://[^ ]+}i,
                title => 1,
                '*' => 0,
            }
        );
    }

    if (&isAffirmative($allowImages, 0)){
        push(@scrubberAllow, "img");
        push(
            @scrubberRules,
            img => {
                src => qr{^http[s]?://[^ ]+}i,
                width => 1,
                height => 1,
                alt => 1,
                '*' => 0,
            }
        );
    }

    my $scrubber = HTML::Scrubber->new(
        allow   => \@scrubberAllow,
        rules   => \@scrubberRules,
        default => 0,
        comment => 0,
        process => 0,
    );
    $s = $scrubber->scrub($s);
    $s =~ s/\n+[ \t\f]*/\n/g;
    $s =~ s/^[\r\n \t\f]*//gs;
    $s =~ s/[\r\n \t\f]*$//gs;
    return $s;
}

sub flush(){
    $| = 1;
    print "";
    $| = 0;
}

sub toInt(){
    my ($s) = @_;
    if ($s and $s =~ /[0-9]+/){
        return int($s);
    }
    return 0;
}

sub thisurl(){
    if ($ENV{'SERVER_NAME'}){
        my $url = "http://";
        $url = $url.$ENV{'SERVER_NAME'};
        if ($ENV{'SERVER_PORT'} and $ENV{'SERVER_PORT'} ne "80"){
            $url=$url.":".$ENV{'SERVER_PORT'};
        }
        if ($ENV{'SCRIPT_NAME'}){
            my $scriptname = $ENV{'SCRIPT_NAME'};
            $scriptname =~ s/index\.cgi//g;
            $url = $url.$scriptname;
        }
        return $url;
    } else {
        return "./";
    }
}

sub alphaNumOnly(){
    my ($s) = @_;
    $s = detaint($s);
    $s =~ s/[^A-Za-z0-9]//g;
    return $s;
}

sub detaint(){
    my ($s) = @_;
    $s = &ensureExists($s);
    $s =~ /(.*)/;
    $s = $1;
    return $s;
}

sub parseDate(){
    my ($date) = @_;
    if (!$date){
        return 0;
    }
    my $t;
    if ($date =~ /^([0-9]{4})\-([0-9]{2})\-([0-9]{2})T([0-9]{2})\:([0-9]{2})\:([0-9]{2})(?:(?:([\+\-][0-9]{2})\:([0-9]{2}))|([a-zA-Z]{1,5}))$/){
        $t = timegm($6,$5,$4,$3,$2-1,$1);
        $t = &correctForAmPmZone($t,0,$7,$8,$9);
    } elsif ($date =~ /^(?:Sun|Mon|Tue|Wed|Thu|Fri|Sat)\, ([0-9]{1,2}) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) ([0-9]{4}) ([0-9]{1,2})\:([0-9]{2})(?:\:([0-9]{2}))? (?:(am|pm|AM|PM) )?(?:(?:([\+\-][0-9]{2})([0-9]{2}))|([a-zA-Z]{1,5}))$/){
        my %mon=("Jan",0,"Feb",1,"Mar",2,"Apr",3,"May",4,"Jun",5,"Jul",6,"Aug",7,"Sep",8,"Oct",9,"Nov",10,"Dec",11);
        $t = timegm(&ensureExistsZero($6),$5,$4,$1,$mon{$2},$3);
        $t = &correctForAmPmZone($t,$7,$8,$9,$10,$4);
    } else {
        print STDERR "Could not parse date: $date\n";
        return 0;
    }
    return $t+0;
}

sub getAgeString(){
    my ($date) = @_;
    if (!$date){
        return "";
    }
    my $age = time() - $date;
    my $ageString = "";    
    my @timeperiodnames=("year","month","week","day","hour","minute","second"); 
    my @timeperiodlengths=(60*60*24*7*4*12,60*60*24*7*4,60*60*24*7,60*60*24,60*60,60,1);
    my $printCount=0;
    
    my $workingAge = abs($age);
    
    for (my $i=0; $i<=$#timeperiodnames; $i++){
        my $val = int($workingAge/$timeperiodlengths[$i]);
        $workingAge = $workingAge%$timeperiodlengths[$i];
        if ($printCount < 2 and ($printCount > 0 or $val != 0)){
            if ($val != 0){
                if ($ageString ne ""){
                    $ageString = $ageString.", ";
                }
                $ageString = $ageString.$val." ".$timeperiodnames[$i];
                if ($val != 1){
                    $ageString=$ageString."s";
                }
            }
            $printCount++;
        }
    }
    if ($ageString eq ""){
        $ageString = "Now";
    } elsif ($age =~ /^-/){
        $ageString = "In ".$ageString;
    } else {
        $ageString = $ageString." ago";
    }
    return $ageString;    
}

sub readXML(){
  my ($file) = @_;
  
  my $fh;
  my $encoding="UTF-8";
  
  open $fh, "<:encoding($encoding)", $file or die "open($file): $!";
  my $line = <$fh>;
  if ($line =~ /encoding\=[\"\']([^\"\']+)[\"\']/i){
    $encoding=$1;
  } else {
    print STDERR "No character encoding found in file: $file\n";
  } 
  close($fh);
  
  open $fh, "<:encoding($encoding)", $file or die "open($file): $!";
  my $contents = join("",<$fh>);
  close($fh);
  
  return $contents;
  
}

sub correctForAmPmZone(){
    my ($t,$ampm,$hours,$minutes,$zone,$hour) = @_;
    my %zones=(
        "idlw","-1200",
        "nt","-1100",
        "hst","-1000",
        "cat","-1000",
        "ahst","-1000",
        "akst","-0900",
        "yst","-0900",
        "hdt","-0900",
        "akdt","-0800",
        "ydt","-0800",
        "pst","-0800",
        "pdt","-0700",
        "mst","-0700",
        "mdt","-0600",
        "cst","-0600",
        "cdt","-0500",
        "est","-0500",
        "act","-0500",
        "sat","-0400",
        "bot","-0400",
        "amt","-0400",
        "acst","-0400",
        "edt","-0400",
        "ast","-0400",
        "nft","-0330",
        "brt","-0300",
        "brst","-0300",
        "adt","-0300",
        "art","-0300",
        "amst","-0300",
        "ndt","-0230",
        "brst","-0200",
        "fnt","-0200",
        "at","-0200",
        "wat","-0100",
        "fnst","-0100",
        "gmt","+0000",
        "ut","+0000",
        "utc","+0000",
        "wet","+0000",
        "cet","+0100",
        "fwt","+0100",
        "met","+0100",
        "mez","+0100",
        "mewt","+0100",
        "swt","+0100",
        "bst","+0100",
        "gb","+0100",
        "west","+0000",
        "eet","+0200",
        "cest","+0200",
        "fst","+0200",
        "ist","+0200",
        "mest","+0200",
        "mesz","+0200",
        "metdst","+0200",
        "sast","+0200",
        "sst","+0200",
        "bt","+0300",
        "eest","+0300",
        "eetedt","+0300",
        "idt","+0300",
        "msk","+0300",
        "eat","+0300",
        "it","+0330",
        "zp4","+0400",
        "msd","+0400",
        "zp5","+0500",
        "ist","+0530",
        "zp6","+0600",
        "novst","+0600",
        "nst","+0630",
        "javt","+0700",
        "hkt","+0800",
        "sgt","+0800",
        "cct","+0800",
        "awst","+0800",
        "wst","+0800",
        "pht","+0800",
        "kst","+0900",
        "jst","+0900",
        "rok","+0900",
        "acst","+0930",
        "cast","+0930",
        "aest","+1000",
        "east","+1000",
        "gst","+1000",
        "acdt","+1030",
        "cadt","+1030",
        "aedt","+1100",
        "eadt","+1100",
        "idle","+1200",
        "nzst","+1200",
        "nzt","+1200",
        "nzdt","+1300",
        "z","+0000",
        "a","+0100",
        "b","+0200",
        "c","+0300",
        "d","+0400",
        "e","+0500",
        "f","+0600",
        "g","+0700",
        "h","+0800",
        "i","+0900",
        "k","+1000",
        "l","+1100",
        "m","+1200",
        "n","-0100",
        "o","-0200",
        "p","-0300",
        "q","-0400",
        "r","-0500",
        "s","-0600",
        "t","-0700",
        "u","-0800",
        "v","-0900",
        "w","-1000",
        "x","-1100",
        "y","-1200",
    );
    if ($zone){
        if ($zones{lc($zone)} && $zones{lc($zone)} =~ /^([\+\-][0-9]{2})([0-9]{2})$/){
            $hours = $1;
            $minutes = $2; 
        } else {
            print STDERR  "Could not parse time zone on date: $zone, ",$zones{lc($zone)},"\n";
        }      
    }
    if (!$minutes){
        $minutes = "0";
    }
    if (!$hours){
        $hours = "0";
    }
    $hours =~ s/\+//g;
    
    $hours=int($hours);
    $minutes=int($minutes);
    if ($ampm){
        if (lc($ampm) eq "pm" and $hour != 12){
            $hours -= 12;
        }
        if (lc($ampm) eq "am" and $hour == 12){
            $hours -= 12;
        }
    }
    
    return $t - (60*60*$hours) - (60*$minutes);
}

sub rssitem(){
    my ($item,$feed,$group,$rss) = @_;
    my $self = {};
    bless $self;
    $self->{ITEM} = $item;
    $self->{FEED} = $feed;
    $self->{GROUP} = $group;
    $self->{RSS} = $rss;
    $self->{TIME} = 0;
    return $self;
}

sub item(){
    my ($self) = @_;
    return $self->{ITEM};
}

sub feed(){
    my ($self) = @_;
    return $self->{FEED};
}

sub rss(){
    my ($self) = @_;
    return $self->{RSS};
}

sub group(){
    my ($self) = @_;
    return $self->{GROUP};
}

sub addDiscussLink(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("addDiscussLink", 0));
}

sub urlInDescription(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("urlInDescription", 0));
}

sub allowFormatting(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("allowFormatting", 0));
}

sub allowLinks(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("allowLinks", 0));
}

sub allowImages(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("allowImages", 0));
}

sub showTitle(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("showTitle", 1));
}

sub showSource(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("showSource", 0));
}

sub showFeedDescription(){
    my ($self) = @_;
    return &isAffirmative($self->getValue("showFeedDescription", 0));
}

sub getValue(){
    my ($self, $name, $default) = @_;
    if ($self->feed()->find('@'.$name)){
        return $self->feed()->find('@'.$name)->string_value();
    }
    if ($self->group() and $self->group()->find('@'.$name)){
        return $self->group()->find('@'.$name)->string_value();
    }
    return $default;
}

sub showDescription(){
    my ($self) = @_;
    return $self->getValue("showDescription", "yes");
}

sub strip(){
    my ($self) = @_;
    return $self->getValue("strip", undef);
}

sub filter(){
    my ($self) = @_;
    return $self->getValue("filter", undef);
}

sub source(){
    my ($self) = @_;
    my $feedTitle = $self->feed()->find('@feedTitle')->string_value();
    if ($feedTitle){
        return $feedTitle;
    }
    return  $self->rss()->channel('title');
}

sub printIndex(){

print <<ZZZHEAD;
<html>
<head>
<title>News RSS</title>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<meta name="description" content="News aggregated from RSS feeds">
</head>
<body>
<h1>RSS Reader</h1>
ZZZHEAD

opendir(DIR, ".") || die "Cannot opendir .: $!";
my @files = sort(grep(/^[A-Za-z]+\.page\.xml$/,readdir(DIR)));
closedir(DIR);
if ($#files < 0){
    print "Error: No files were found\n";
} else {
    print "<h2>Feed Groups</h2><ul>\n";
    foreach my $file (@files){
        my $name = $file;
        $name =~ s/\.page\.xml$//g;
        my $title = $name;
        my $description = "";
        open (FILE, $file) or die "could not open $file: $!";
            my $line = <FILE>;
            if ($line =~ /\<page .*title\=\"([^\"]+)\"/){
                $title = $1;
            }
            if ($line =~ /\<page .*description\=\"([^\"]+)\"/){
                $description = $1;
            }
        close FILE;
        if ($description){
            $description = " -- $description";
        }
        print "<li><a href=\"$name.html\">$title</a>$description</li>\n";

    }
    print "</ul>\n";
}


print <<ZZZFOOT;
<p align=right>Powered By: <a href="http://ostermiller.org/rssreader/">Ostermiller RSS Reader</a></p>
</body>
</html>
ZZZFOOT


}

